# python -m unittest discover -s tests -t .

import unittest

from muxllm import LLM, Provider
from muxllm.tools import tool, Param, ToolBox
from muxllm.providers.base import ToolCall

my_tools = ToolBox()

@tool("get_current_weather", my_tools, "Get the current weather", [
    Param("location", "string", "The city and state, e.g. San Francisco, CA"),
    Param("format", "string", "The temperature unit to use. Infer this from the users location.", enum=["celsius", "fahrenheit"])
])
def get_current_weather(location, format):
    return f"It is sunny in {location} according to the weather forecast in {format}"

@tool("get_n_day_weather_forecast", my_tools, "Get an N-day weather forecast", [
    Param("location", "string", "The city and state, e.g. San Francisco, CA"),
    Param("format", "string", "The temperature unit to use. Infer this from the users location.", enum=["celsius", "fahrenheit"]),
    Param("num_days", "integer", "The number of days to forecast")
])
def get_n_day_weather_forecast(location, format, num_days):
    return f"It will be sunny for the next {num_days} days in {location} according to the weather forecast in {format}"

TEST_TOOLS = [
    {
        "type": "function",
        "function": {
            "name": "get_current_weather",
            "description": "Get the current weather",
            "parameters": {
                "type": "object",
                "properties": {
                    "location": {
                        "type": "string",
                        "description": "The city and state, e.g. San Francisco, CA",
                    },
                    "format": {
                        "type": "string",
                        "enum": ["celsius", "fahrenheit"],
                        "description": "The temperature unit to use. Infer this from the users location.",
                    },
                },
                "required": ["location", "format"],
            },
        }
    },
    {
        "type": "function",
        "function": {
            "name": "get_n_day_weather_forecast",
            "description": "Get an N-day weather forecast",
            "parameters": {
                "type": "object",
                "properties": {
                    "location": {
                        "type": "string",
                        "description": "The city and state, e.g. San Francisco, CA",
                    },
                    "format": {
                        "type": "string",
                        "enum": ["celsius", "fahrenheit"],
                        "description": "The temperature unit to use. Infer this from the users location.",
                    },
                    "num_days": {
                        "type": "integer",
                        "description": "The number of days to forecast",
                    }
                },
                "required": ["location", "format", "num_days"]
            },
        }
    },
]

class TestTools(unittest.TestCase):
    def test_toolbox(self):
        self.assertEqual(my_tools.to_dict(), TEST_TOOLS)
        self.assertEqual(my_tools.invoke_tool(
            ToolCall(id="1", name="get_current_weather", args={"location": "San Francisco, CA", "format": "fahrenheit"})
        ), "It is sunny in San Francisco, CA according to the weather forecast in fahrenheit")

    def test_openai_tools(self):
        llm = LLM(Provider.openai, "gpt-4-turbo")
        response = llm.chat("What is the weather in San Francisco, CA in fahrenheit", tools=TEST_TOOLS)
        self.assertNotEqual(response.tools, None)
        self.assertEqual(response.tools[0].name, "get_current_weather")

        self.assertEqual(my_tools.invoke_tool(response.tools[0]), "It is sunny in San Francisco, CA according to the weather forecast in fahrenheit")

        llm.add_tool_response(response.tools[0], "It is sunny in San Francisco, CA")
        response = llm.chat("Please tell me what the tool said")
        self.assertTrue("sunny" in response.message.lower())

    def test_fireworks_tools(self):
        llm = LLM(Provider.fireworks, "firefunction-v2")
        response = llm.chat("What is the weather in San Francisco, CA", tools=TEST_TOOLS)
        self.assertNotEqual(response.tools, None)
        self.assertEqual(response.tools[0].name, "get_current_weather")

        llm.add_tool_response(response.tools[0], "It is sunny in San Francisco, CA")
        response = llm.chat("Please tell me what the tool said")
        self.assertTrue("sunny" in response.message.lower())

    def test_anthropic_tools(self):
        llm = LLM(Provider.anthropic, "claude-3-5-sonnet")
        response = llm.chat("What is the weather in San Francisco, CA", tools=TEST_TOOLS, max_tokens=500)
        self.assertNotEqual(response.tools, None)
        self.assertEqual(response.tools[0].name, "get_current_weather")

        llm.add_tool_response(response.tools[0], "It is sunny in San Francisco, CA")
        response = llm.chat("Please tell me what the tool said", max_tokens=500)
        self.assertTrue("sunny" in response.message.lower())

    def test_google_tools(self):
        llm = LLM(Provider.google, "gemini-1.5-pro")
        response = llm.chat("What is the weather in San Francisco, CA", tools=TEST_TOOLS)
        self.assertNotEqual(response.tools, None)
        self.assertEqual(response.tools[0].name, "get_current_weather")

        llm.add_tool_response(response.tools[0], "It is sunny in San Francisco, CA")
        response = llm.chat("Please tell me what the tool said")
        self.assertTrue("sunny" in response.message.lower())


if __name__ == '__main__':
    unittest.main()


import openai, os
from typing import Optional
from muxllm.providers.base import CloudProvider

model_alias = {
    "gpt-4-turbo" : "gpt-4-turbo-preview",
    "gpt-4-vision" : "gpt-4-vision-preview",
}

class BaseOpenAIProvider(CloudProvider):
    def __init__(self, model_alias : dict, base_url : str, api_key : Optional[str] = None):
        super().__init__(model_alias)

        self.client = openai.Client(base_url=base_url, api_key=api_key)
        self.async_client = openai.AsyncClient(base_url=base_url, api_key=api_key)

        self.client.chat.completions.create

class OpenAIProvider(BaseOpenAIProvider):
    def __init__(self, api_key : Optional[str] = None):
        if api_key is None:
            api_key = os.getenv("OPENAI_API_KEY")
        super().__init__(model_alias, base_url="https://api.openai.com/v1", api_key=api_key)



from typing import Any
from muxllm.providers.base import ToolCall

class ToolBox:
    def __init__(self):
        self.tools = {}

    def add_tool(self, tool):
        self.tools[tool.name] = tool

    def get_tool(self, name: str):
        return self.tools.get(name)
    
    def invoke_tool(self, tool_call: ToolCall):
        tool = self.get_tool(tool_call.name)
        if tool:
            return tool(**tool_call.args)
        else:
            return None
        
    def to_dict(self) -> dict[str, Any]:
        return [tool.to_dict() for tool in self.tools.values()]
    
    def __add__(self, other):
        new_toolbox = ToolBox()
        new_toolbox.tools = {**self.tools, **other.tools}
        return new_toolbox

class Param:
    def __init__(self, name: str, type: str, description: str, enum: list[str] = None):
        self.name = name
        self.description = description
        self.type = type
        self.enum = enum

    def to_dict(self) -> dict[str, Any]:
        d = {
            "type": self.type,
            "description": self.description
        }

        if self.enum:
            d["enum"] = self.enum

        return d

class Tool:
    def __init__(self, name: str, description: str, parameters: dict, function: callable):
        self.name = name
        self.description = description
        self.parameters = parameters
        self.function = function

    def to_dict(self) -> dict[str, Any]:
        return {
            "type": "function",
            "function" : {
                "name": self.name,
                "description": self.description,
                "parameters": {
                    "type": "object",
                    "properties": {param.name: param.to_dict() for param in self.parameters},
                    "required": [param.name for param in self.parameters]
                }
            }
        }
    
    def __call__(self, *args, **kwargs):
        return self.function(*args, **kwargs)

def tool(name: str,
         toolBox: ToolBox,
         description: str,
         parameters: list[Param]):
    def decorator(func):
        tool = Tool(name, description, parameters, func)
        toolBox.add_tool(tool)
        return func
    return decorator


from groq import Groq, AsyncGroq
from typing import Optional
import os

from muxllm.providers.base import CloudProvider

model_alias = {
    "llama3-8b-instruct": "llama3-8b-8192",
    "llama3-70b-instruct": "llama3-70b-8192",
    "mixtral-8x7b-instruct": "mixtral-8x7b-32768",
    "gemma-7b-instruct": "gemma-7b-it",
    "gemma2-9b-instruct": "gemma-9b-it",
}

class GroqProvider(CloudProvider):
    def __init__(self, api_key : Optional[str] = None):
        super().__init__(model_alias)
        if api_key is None:
            api_key = os.getenv("GROQ_API_KEY")
        self.client = Groq(api_key=api_key)
        self.async_client = AsyncGroq(api_key=api_key)


import google.generativeai as genai
import proto
import os
from muxllm.providers.base import CloudProvider, LLMResponse, ToolCall, ToolResponse
from typing import Optional

model_alias = {}

class GoogleProvider(CloudProvider):
    def __init__(self, api_key : Optional[str] = None):
        super().__init__(model_alias)
        if api_key is None:
            api_key = os.getenv("GOOGLE_API_KEY")

        genai.configure(api_key=api_key)

    def parse_system_message(self, message: str) -> dict:
        # TODO: 
        return {
            "role": "system",
            "parts": [message]
        }

    def parse_user_message(self, message: str) -> dict:
        return {
            "role": "user",
            "parts": [message]
        }

    def parse_response(self, response: LLMResponse) -> dict:
        if not response.tools:
            resp = {"role": "model",
                    "parts": [genai.protos.Part({"text": response.message})] if response.message else []}
        else:
            resp = {"role": "model",
                    "parts": [genai.protos.Part({"function_call": genai.protos.FunctionCall({  
                        "name": tool.name,
                        "args": tool.args
                        })}) for tool in response.tools]
                    }
        return genai.protos.Content(resp)



    def parse_tool_response(self, tool_resp: ToolResponse) -> dict:
        return genai.protos.Content({
            "role": "function",
            "parts": [{
                "function_response": genai.protos.FunctionResponse({
                    "name": tool_resp.name,
                    "response": {
                        "name": tool_resp.name,
                        "content": tool_resp.response
                    }
            })}]
        })

    def tools_dict_to_google_protos(self, tools: list[dict[str, str | dict]]) -> list[genai.protos.Tool]:
        google_proto_tools = []
        for tool in tools:
            google_proto_tool = {
                'function_declarations': [
                    {
                        'name': tool['function']['name'],
                        'description': tool['function']['description'],
                        'parameters': {
                            'type_': tool['function']['parameters']['type'].upper(),
                            'properties': {
                                prop_name: {'type_': prop_data['type'].upper(), "description": prop_data['description']} for prop_name, prop_data in tool['function']['parameters']['properties'].items()
                            },
                            'required': tool['function']['parameters']['required']
                        }
                    }
                ]
            }
            google_proto_tools.append(genai.protos.Tool(google_proto_tool))
        return google_proto_tools

    def get_response(self, messages : list[dict[str, str | dict]], model : str, **kwargs) -> LLMResponse:
        model = self.validate_model(model)

        google_proto_tools = []
        if "tools" in kwargs:
            google_proto_tools = self.tools_dict_to_google_protos(kwargs["tools"])
        # google doesnt need tool_choice, delete it if it exists
        if "tool_choice" in kwargs:
            del kwargs["tool_choice"]

        if messages[0]["role"] == "system":
            system_message = messages[0]["parts"][0]
            client = genai.GenerativeModel(model, system_instruction=system_message, tools=google_proto_tools)
            messages = messages[1:]
        else:
            client = genai.GenerativeModel(model, tools=google_proto_tools)

        response = client.generate_content(messages)

        
        tools = []
        for part in response.candidates[0].content.parts:
            if fn := part.function_call:
                tools.append(ToolCall(id='', name=fn.name, args={k: v for k, v in fn.args.items()}))
        if tools:
            return LLMResponse(model=model, raw_response=response, message="", tools=tools)
        else:
            return LLMResponse(model=model, raw_response=response, message=response.text, tools=None)
    
    async def get_response_async(self, messages : list[dict[str, str | dict]], model : str, **kwargs) -> LLMResponse:
        model = self.validate_model(model)

        google_proto_tools = []
        if "tools" in kwargs:
            google_proto_tools = self.tools_dict_to_google_protos(kwargs["tools"])

        if messages[0]["role"] == "system":
            system_message = messages[0]["parts"][0]
            client = genai.GenerativeModel(model, system_instruction=system_message, tools=google_proto_tools)
            messages = messages[1:]
        else:
            client = genai.GenerativeModel(model, tools=google_proto_tools)

        response = await client.generate_content_async(messages)
        
        tools = []
        for part in response.candidates[0].content.parts:
            if fn := part.function_call:
                tools.append(ToolCall(id='', name=fn.name, args={k: v for k, v in fn.args.items()}))
        if tools:
            return LLMResponse(model=model, raw_response=response, message="", tools=tools)
        else:
            return LLMResponse(model=model, raw_response=response, message=response.text, tools=None)


# python -m unittest discover -s tests -t .

import unittest

from muxllm import LLM, Prompt, Provider
from muxllm.llm import SinglePromptLLM
import os


class TestLLM(unittest.TestCase):
    def test_history(self):
        llm = LLM(Provider.openai, "gpt-3.5-turbo", system_prompt="You are a helpful assistant that answers questions")
        llm.chat("What is the capital of France?")
        llm.save_history("history.json")
        llm.load_history("history.json")
        self.assertEqual(llm.history[1]["content"], "What is the capital of France?")
        self.assertEqual(llm.history[2]["role"], "assistant")
        os.remove("history.json")

    def test_prompts(self):
        prompt = Prompt("My very cool prompt with {{default}} argument, {{changing}} argument, and {{another}} argument", default="im setting this default")
        self.assertEqual(prompt.get(changing="this is changing", another="another"), "My very cool prompt with im setting this default argument, this is changing argument, and another argument")
        self.assertEqual(prompt.get(), "My very cool prompt with im setting this default argument, {{changing}} argument, and {{another}} argument")

    def test_ask(self):
        llm = LLM(Provider.openai, "gpt-3.5-turbo")
        response = llm.ask("Translate {{spanish}} to english", spanish="Hola, como estas?").message
        self.assertEqual("Hello, how are you?" in response, True)

    def test_single_prompt(self):
        llm = SinglePromptLLM(Provider.fireworks, "llama3-8b-instruct", "Translate {{spanish}} to english", system_prompt="You are a helpful translator that translates from spanish to english")
        response = llm.ask(spanish="Hola, como estas?").message
        self.assertEqual("Hello, how are you?" in response, True)



from pydantic import BaseModel
import json

class ModelNotAvailable(Exception):
    pass

class ToolCall(BaseModel):
    id: str
    name: str
    args: dict[str, str]

class ToolResponse(BaseModel):
    id: str
    name: str
    response: str

class LLMResponse(BaseModel):
    model: str
    raw_response: dict | object # raw response from the provider. Ideally dict, but google uses protobuf natively
    message: str | None
    tools: list[ToolCall] | None

class BaseProvider:
    def __init__(self):
        pass

    def parse_system_message(self, message : str) -> dict:
        pass

    def parse_user_message(self, message : str) -> dict:
        pass

    def parse_response(self, response: LLMResponse) -> dict:
        pass

    def parse_tool_response(self, tool_resp: ToolResponse) -> dict:
        pass

    def get_response(self, messages : list[dict[str, str | dict]], model : str, **kwargs) -> LLMResponse:
        pass

    async def get_response_async(self, messages : list[dict[str, str | dict]], model : str, **kwargs) -> LLMResponse:
        pass

    # def get_response_stream(self, messages : list[dict[str, str]], model : str, **kwargs):
    #     pass

class CloudProvider(BaseProvider):
    def __init__(self, model_alias : dict[str, str]):
        self.model_alias = model_alias
        self.client = None
        self.async_client = None

    def validate_model(self, model : str): 
        if model in self.model_alias:
            model = self.model_alias[model]
        return model
    
    def parse_system_message(self, message : str) -> dict:
        return {
            "role": "system",
            "content": message
        }
    
    def parse_user_message(self, message : str) -> dict:
        return {
            "role": "user",
            "content": message
        }
    
    def parse_response(self, response: LLMResponse) -> dict:
        msg = {
            "role": "assistant",
            "content": response.message if response.message else "",
        }
        if response.tools:
            msg["tool_calls"] = [
                {
                    "id": tool.id,
                    "type": "function",
                    "function": {
                        "name": tool.name,
                        "arguments": json.dumps(tool.args)
                    }
                } for tool in response.tools if response.tools is not None
            ]

        return msg
    
    def parse_tool_response(self, tool_resp: ToolResponse) -> dict:
        return {
            "role": "tool",
            "tool_call_id": tool_resp.id,
            "name": tool_resp.name,
            "content": tool_resp.response
        }

    def get_response(self, messages : list[dict[str, str | dict]], model : str, **kwargs) -> LLMResponse:
        model = self.validate_model(model)
        
        response = self.client.chat.completions.create(
                    model=model,
                    messages=messages,
                    **kwargs) 
        message = response.choices[0].message

        resp = LLMResponse(model=model, raw_response=dict(response), message=message.content, tools=[
                            ToolCall(id=message.tool_calls[i].id, name=message.tool_calls[i].function.name, args=json.loads(message.tool_calls[i].function.arguments))
                                for i in range(len(message.tool_calls))] if message.tool_calls else None)
        return resp
    
    async def get_response_async(self, messages : list[dict[str, str | dict]], model : str, **kwargs) -> LLMResponse:
        model = self.validate_model(model)

        response = await self.client.chat.completions.create(
                    model=model,
                    messages=messages,
                    **kwargs) 
        message = response.choices[0].message
        resp = LLMResponse(model=model, raw_response=dict(response), message=message.content, tools=[
                    ToolCall(id=message.tool_calls[i].id, name=message.tool_calls[i].function.name, args=json.loads(message.tool_calls[i].function.arguments))
                        for i in range(len(message.tool_calls))] if message.tool_calls else None)
        return resp
    
    # def get_response_stream(self, messages : list[dict[str, str]], model : str, **kwargs):
    #     model = self.validate_model(model)
        
    #     response = self.client.chat.completions.create(
    #                 model=model,
    #                 messages=messages,
    #                 **kwargs) 
        
    #     for chunk in response:
    #         yield LLMResponse(model=model, raw_response=chunk, message=chunk.choices[0].delta.content, tool=chunk.choices[0].delta.tools)





from muxllm.providers.popenai import BaseOpenAIProvider
from typing import Optional
import os

model_alias = {
    "firefunction-v2": "accounts/fireworks/models/firefunction-v2",
    "mixtral-8x7b-instruct": "accounts/fireworks/models/mixtral-8x7b-instruct",
    "mixtral-8x22b-instruct": "accounts/fireworks/models/mixtral-8x22b-instruct",
    "llama3-8b-instruct": "accounts/fireworks/models/llama-v3-8b-instruct",
    "llama3-70b-instruct": "accounts/fireworks/models/llama-v3-70b-instruct",
    "gemma-7b-instruct": "accounts/fireworks/models/gemma-7b-it",
    "gemma2-9b-instruct": "accounts/fireworks/models/gemma-9b-it",
}

available_models = [] # empty means that all models are available, mostly because there are far too many models on fireworks

class FireworksProvider(BaseOpenAIProvider):
    def __init__(self, api_key: Optional[str] = None):
        if api_key is None:
            api_key = os.getenv("FIREWORKS_API_KEY")
        super().__init__(model_alias, base_url="https://api.fireworks.ai/inference/v1", api_key=api_key)


# python -m unittest discover -s tests -t .

import unittest

from muxllm import LLM, Provider

class TestProvider(unittest.TestCase):
    def test_openai_provider(self):
        llm = LLM(Provider.openai, "gpt-3.5-turbo")
        response = llm.ask("Translate 'Hola, como estas?' to english")
        self.assertEqual("how are you?" in response.message, True)

    def test_groq_provider(self):
        llm = LLM(Provider.groq, "llama3-8b-instruct")
        response = llm.ask("Translate 'Hola, como estas?' to english")
        self.assertEqual("how are you?" in response.message, True)

    def test_fireworks_provider(self):
        llm = LLM(Provider.fireworks, "llama3-8b-instruct")
        response = llm.ask("Translate 'Hola, como estas?' to english")
        self.assertEqual("how are you?" in response.message, True)

    def test_anthropic_provider(self):
        llm = LLM(Provider.anthropic, "claude-3-5-sonnet")
        response = llm.ask("Translate 'Hola, como estas?' to english", max_tokens=500)
        self.assertEqual("how are you?" in response.message, True)

    def test_google_provider(self):
        llm = LLM(Provider.google, "gemini-1.5-pro")
        response = llm.ask("Translate 'Hola, como estas?' to english")
        print(response.message)
        self.assertEqual("how are you?" in response.message, True)

if __name__ == '__main__':
    unittest.main()


import os
from typing import Optional
from muxllm.providers.base import CloudProvider, LLMResponse, ToolCall, ToolResponse
import anthropic

model_alias = {
    "claude-3-5-sonnet": "claude-3-5-sonnet-20240620",
    "claude-3-opus": "claude-3-opus-20240229",
    "claude-3-sonnet": "claude-3-sonnet-20240229",
    "claude-3-haiku": "claude-3-haiku-20240307",
}

class AnthropicProvider(CloudProvider):
    def __init__(self, api_key : Optional[str] = None):
        super().__init__(model_alias)
        if api_key is None:
            api_key = os.getenv("ANTHROPIC_API_KEY")
        self.client = anthropic.Anthropic(api_key=api_key)
        self.async_client = anthropic.AsyncAnthropic(api_key=api_key)

    def parse_response(self, response: LLMResponse) -> dict:
        if not response.tools:
            return {"role": "assistant",
                    "content": response.message if response.message else ""}
        else:
            return {"role": "assistant",
                    "content": [{"type": " text",
                                 "text" : response.message if response.message else ""}] + 
                               [{"type": "tool_use",
                                 "id": tool.id,
                                 "name": tool.name,
                                 "input": tool.args
                               } for tool in response.tools]}  

    def parse_tool_response(self, tool_resp: ToolResponse) -> dict:
        return {
            "role": "user",
            "content": {
                "type": "tool_result",
                "tool_use_id": tool_resp.id,
                "content": tool_resp.response
            }
        }

    def get_response(self, messages : list[dict[str, str | dict]], model : str, **kwargs) -> LLMResponse:
        model = self.validate_model(model)
        
        response = self.client.messages.create(
                    model=model,
                    messages=messages,
                    **kwargs) 
        
        if response.stop_reason == "tool_use":
            tool_uses = [block for block in response.content if block.type == "tool_use"]
            thinking = next(block for block in response.content if block.type == "text")
            return LLMResponse(model=model, raw_response=dict(response), message=thinking.text, tools=[ToolCall(id=tool_use.id, name=tool_use.name, args=tool_use.input) for tool_use in tool_uses])
        return LLMResponse(model=model, raw_response=response, message=response.content.text, tools=None)
    
    async def get_response_async(self, messages : list[dict[str, str | dict]], model : str, **kwargs) -> LLMResponse:
        model = self.validate_model(model)

        response = await self.async_client.messages.create(
                            model=model,
                            messages=messages,
                            **kwargs) 
        
        if response.stop_reason == "tool_use":
            tool_uses = [block for block in response.content if block.type == "tool_use"]
            thinking = next(block for block in response.content if block.type == "text")
            return LLMResponse(model=model, raw_response=dict(response), message=thinking.text, tools=[ToolCall(id=tool_use.id, name=tool_use.name, args=tool_use.input) for tool_use in tool_uses])
        return LLMResponse(model=model, raw_response=response, message=response.content.text, tools=None)
    
    
    # def get_response_stream(self, messages : list[dict[str, str]], model : str, **kwargs):
    #     model = self.validate_model(model)
        
    #     response = self.client.messages.stream(
    #                 model=model,
    #                 messages=messages,
    #                 **kwargs) 
        
    #     for chunk in response:
    #         yield LLMResponse(model=model, raw_response=chunk, message=chunk.content.text, tools=chunk.choices[0].message.tools)


import os

class Prompt:
    def __init__(self, prompt, **kwargs):
        if os.path.exists("./prompts/" + prompt):
            with open("./prompts/" + prompt, "r") as f:
                self.raw_prompt = f.read()
        elif os.path.exists(prompt):
            with open(prompt, "r") as f:
                self.raw_prompt = f.read()
        else:
            self.raw_prompt = prompt
        
        self.raw_prompt, _ = self.prep_prompt(self.raw_prompt, **kwargs)

    def __str__(self) -> str:
        return self.raw_prompt

    def prep_prompt(self, prompt, **kwargs):
        new_kwargs = {}
        for key, value in kwargs.items():
            key_bracketed = "{{" + str(key) + "}}"
            if not key_bracketed in prompt:
                new_kwargs[key] = value
            else:
                prompt = prompt.replace(key_bracketed, value)
        return prompt, new_kwargs

    def get(self, **kwargs):
        prompt, _ = self.prep_prompt(self.raw_prompt, **kwargs)
        return prompt
        
    def get_kwargs(self, **kwargs):
        return self.prep_prompt(self.raw_prompt, **kwargs)

            
    



from enum import Enum
from muxllm.providers import pfireworks, popenai, pgroq, panthropic, pgoogle
import importlib # for local provider
from muxllm.providers.base import CloudProvider

# create an enum for the available providers
class Provider(str, Enum):
    openai = "openai"
    groq = "groq"
    fireworks = "fireworks"
    anthropic = "anthropic"
    google = "google"
    local = "local"

# create a factory method to create the correct provider
def create_provider(provider: Provider, api_key=None) -> CloudProvider:
    if provider == Provider.openai:
        return popenai.OpenAIProvider(api_key)
    elif provider == Provider.groq:
        return pgroq.GroqProvider(api_key)
    elif provider == Provider.fireworks:
        return pfireworks.FireworksProvider(api_key)
    elif provider == Provider.anthropic:
        return panthropic.AnthropicProvider(api_key)
    elif provider == Provider.google:
        return pgoogle.GoogleProvider(api_key)
    elif provider == Provider.local:
        try:
            import llama_cpp
        except:
            raise ValueError("Local provider requires the llama_cpp package to be installed")
        return importlib.import_module("muxllm.providers.plocal").LocalProvider()
    else:
        raise ValueError(f"Provider {provider} is not available")


from .providers.factory import Provider, create_provider
from .providers.base import ToolCall, ToolResponse, LLMResponse
from .tools import ToolBox
from .prompt import Prompt
from typing import Optional, Union
import json

'''
# usage

llm = LLM(Provider.groq, "llama3-8b-instruct", api_key="your-api-key", system_prompt="You are a helpful translator that translates from spanish to english")
english = llm.ask("Translate "Hola, como estas?" to english")

# example with single prompt

llm = SinglePromptLLM(Provider.groq, "llama3-8b-instruct", "Translate {{spanish}} to english", system_prompt="You are a helpful translator that translates from spanish to english")

english = llm.ask(spanish="Hola, como estas?")

# example with chat and function calling

llm = LLM(Provider.openai, "gpt-4")

resp = llm.chat("Search up the weather in New York", tools=[...])

# call tool somehow
tool_response = call_tool(resp.tools[0])

# add tool response to history to keep track of it
llm.add_tool_response(resp.tools[0], tool_response)

'''

class LLM:
    def __init__(self, provider: Provider, model : str,  api_key : Optional[str] = None, system_prompt : Optional[Union[str, Prompt]] = None):
        self.provider = create_provider(provider, api_key)
        self.model = model
        self.system_prompt = system_prompt
        self.history = []

        if system_prompt is not None:
            self.history.append(self.provider.parse_system_message(system_prompt))

    def __call__(self, messages: list, **kwargs):
        return self.provider.get_response(messages, self.model, **kwargs)

    def save_history(self, fp : str):
        with open(fp, "w") as f:
            json.dump(self.history, f, indent=4)

    def load_history(self, fp : str):
        with open(fp, "r") as f:
            self.history = json.load(f)

    def reset(self):
        self.history = []

    def prep_prompt(self, prompt : Union[str, Prompt], **kwargs):
        if isinstance(prompt, str):
            prompt = Prompt(prompt)
        return prompt.get_kwargs(**kwargs)

    def ask(self, prompt: Union[str, Prompt], system_prompt : Optional[Union[str, Prompt]] = None, **kwargs) -> LLMResponse:
        prompt, kwargs = self.prep_prompt(prompt, **kwargs)

        # if tools is in kwargs, check if its a ToolBox and convert it to a dict
        if "tools" in kwargs:
            tools = kwargs["tools"]
            if isinstance(tools, ToolBox):
                kwargs["tools"] = tools.to_dict()

        messages = []

        if system_prompt:
            system_prompt, kwargs = self.prep_prompt(system_prompt, **kwargs)
            messages.append(self.provider.parse_system_message(system_prompt))
        elif self.system_prompt:
            messages.append(self.provider.parse_system_message(self.system_prompt))

        messages.append(self.provider.parse_user_message(prompt))

        response = self.provider.get_response(messages, self.model, **kwargs)

        return response
     
    def chat(self, prompt: Union[str, Prompt], **kwargs) -> LLMResponse:
        prompt, kwargs = self.prep_prompt(prompt, **kwargs)

        if "tools" in kwargs:
            tools = kwargs["tools"]
            if isinstance(tools, ToolBox):
                kwargs["tools"] = tools.to_dict()

        self.history.append(self.provider.parse_user_message(prompt))

        response = self.provider.get_response(self.history, self.model, **kwargs)

        self.history.append(self.provider.parse_response(response))

        return response
    
    def add_user_message(self, message: str):
        self.history.append(self.provider.parse_user_message(message))

    def add_model_message(self, message: str):
        self.history.append(self.provider.parse_response(message))
    
    def add_tool_response(self, tool_call: ToolCall, tool_response: str):
        tool_response = ToolResponse(id=tool_call.id, name=tool_call.name, response=tool_response)
        self.history.append(self.provider.parse_tool_response(tool_response))

class SinglePromptLLM(LLM):
    def __init__(self, provider: Provider, model : str, prompt : Union[str, Prompt], system_prompt : Optional[Union[str, Prompt]] = None, api_key : Optional[str] = None, **kwargs):
        super().__init__(provider, model, api_key=api_key, system_prompt=system_prompt)
        if isinstance(prompt, Prompt):
            prompt = prompt.get(**kwargs)
        self.prompt = prompt

    def ask(self, **kwargs):
        return super().ask(self.prompt, **kwargs)
